/*
 * Copyright (c) Members of the EGEE Collaboration. 2004.
 * See http://www.eu-egee.org/partners/ for details on the copyright holders.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *     http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 * Authors: Zsolt Molnar <zsolt.molnar@cern.ch, http://www.zsoltmolnar.hu>
 *
 * Objects for dependency injection of SRM services.
 * (see: http://en.wikipedia.org/wiki/Dependency_injection)
 *
 * The goal: make the LCG_Util logic unit-testable by decoupling from the real GFAL
 * calls. Unit tests then create mock objects and create response objects
 * without real soap call / network operations.
 */

/*
 * @(#)$RCSfile: gfal_dependencies.h,v $ $Revision: 1.3 $ $Date: 2010/01/21 08:44:26 $ CERN Remi Mollon
 */
#ifndef GFAL_DEPENDENCIES_H__
#define GFAL_DEPENDENCIES_H__

#include "gfal_api.h"
#include "gfal_internals.h"

/* enforce proper calling convention */
#ifdef __cplusplus
extern "C"
{
#endif

/* Function types of the appropriate SOAP calls */
typedef int (*lcg_gfal_turlsfromsurls_fv)(
    gfal_internal req, char *errbuf, int errbufsz);

typedef int (*lcg_gfal_ls_fv)(
    gfal_internal req, char *errbuf, int errbufsz);

typedef int (*lcg_gfal_get_results_fv)(
    gfal_internal req, gfal_filestatus **results);

typedef int (*lcg_gfal_set_xfer_done_fv)(
    gfal_internal gfal, char *errbuf, int errbufsz);

typedef int (*lcg_gfal_register_file_fv) (
    const char *lfn, const char *guid, const char *surl, mode_t mode,
    GFAL_LONG64 size, int bool_createonly, char *errbuf, int errbufsz);

typedef char * (*lcg_gfal_guidforpfn_fv) (
    const char *pfn, char *errbuf, int errbufsz);

typedef int (*lcg_gfal_file_next_replica_fv) (
    gfal_file gf);

typedef gfal_file (*lcg_gfal_file_new_fv) (
    const char *file, const char *defproto, int bool_tobecreated, char *errbuf,
    int errbufsz);

typedef int (*lcg_gfal_init) (
	gfal_request req, gfal_internal *gfal, char *errbuf, int errbufsz);

typedef int (*lcg_gfal_deletesurls) (
	gfal_internal req, char *errbuf, int errbufsz);

/* The collection of GFAL calls. The pointers may be replaced with test
 * functions (dependency injection) */
typedef struct lcg_gfal_callers {
    lcg_gfal_turlsfromsurls_fv call_gfal_turlsfromsurls;
    lcg_gfal_ls_fv call_gfal_ls;
    lcg_gfal_get_results_fv call_gfal_get_results;
    lcg_gfal_set_xfer_done_fv call_gfal_set_xfer_done;
    lcg_gfal_register_file_fv call_gfal_register_file;
    lcg_gfal_guidforpfn_fv call_gfal_guidforpfn;
    lcg_gfal_file_next_replica_fv call_gfal_file_next_replica;
    lcg_gfal_file_new_fv call_gfal_file_new;
    lcg_gfal_init call_gfal_init;
    lcg_gfal_deletesurls call_gfal_deletesurls;
} lcg_gfal_callers_t;

extern lcg_gfal_callers_t lcg_gfal_functions;

#ifdef __cplusplus
}
#endif

#endif /* #ifndef GFAL_DEPENDENCIES_H__ */
