/*
 * Copyright (c) Members of the EGEE Collaboration. 2004.
 * See http://www.eu-egee.org/partners/ for details on the copyright holders.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *     http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#ifndef lint
static char sccsid[] = "@(#)$RCSfile: lcg-gt.c,v $ $Revision: 1.21 $ $Date: 2008/11/21 15:19:48 $ CERN Jean-Philippe Baud";
#endif /* not lint */

/*	lcg-gt - get the TURL for a given SURL and transfer protocol */
#include <sys/types.h>
#include <errno.h>
#include <getopt.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <strings.h>
#include "lcg_util.h"

	int
main (int argc, char **argv)
{
	int helpflg = 0;
	int insecure = 0;
	int verbose = 0;
	int version = 0;
	int c = 0;
	int errflg = 0;
	int fileid;
	int timeout = 0, timeout_tmp = 0;
	static struct option longopts[] = {
		{"help", no_argument, NULL, 'h'},
		{"insecure", no_argument, NULL, 'i'},
		{"st", required_argument, NULL, 's'},
		{"defaultsetype", required_argument, NULL, 'D'},
		{"setype", required_argument, NULL, 'T'},
		{"nobdii", no_argument, NULL, 'b'},
		{"timeout", required_argument, NULL, 't'},
		{"connect-timeout", required_argument, NULL, 130},
		{"sendreceive-timeout", required_argument, NULL, 131},
		{"bdii-timeout", required_argument, NULL, 132},
		{"srm-timeout", required_argument, NULL, 133},
		{"verbose", no_argument, NULL, 'v'},
		{"version", no_argument, NULL, 'W'},
		{0, 0, 0, 0}
	};
	int reqid;
	char *spacetokendesc = NULL;
	char *token;
	char *turl;
	char *dp;
	enum se_type defaulttype = TYPE_NONE, setype = TYPE_NONE, usedtype = TYPE_NONE;
	int nobdii = 0;

	while ((c = getopt_long (argc, argv, "hivt:s:D:T:bW", longopts, NULL)) != EOF) {
		switch (c) {
			case 'h':
				helpflg = 1;
				break;
			case 'i':
				insecure = 1;
				break;
			case 't':
				fprintf (stderr, "Warning: -t,--timeout is deprecated! Use --timeout-* options instead\n");
				if ((timeout = strtol (optarg, &dp, 10)) < 0 || 
						*dp != '\0' ) {
					fprintf (stderr,
							"invalid value for '--timeout': %s\n", optarg);
					errflg++;
				}
			case 130: // "connect-timeout"
				if ((timeout_tmp = strtol (optarg, &dp, 10)) < 0 || 
						*dp != '\0' ) {
					fprintf (stderr,
							"invalid value for '--connect-timeout': %s\n", optarg);
					errflg++;
				} else
					gfal_set_timeout_connect (timeout_tmp);
				break;
			case 131: // "sendreceive-timeout"
				if ((timeout_tmp = strtol (optarg, &dp, 10)) < 0 || 
						*dp != '\0' ) {
					fprintf (stderr,
							"invalid value for '--sendreceive-timeout': %s\n", optarg);
					errflg++;
				} else
					gfal_set_timeout_sendreceive (timeout_tmp);
				break;
			case 132: // "bdii-timeout"
				if ((timeout_tmp = strtol (optarg, &dp, 10)) < 0 || 
						*dp != '\0' ) {
					fprintf (stderr,
							"invalid value for '--bdii-timeout': %s\n", optarg);
					errflg++;
				} else
					gfal_set_timeout_bdii (timeout_tmp);
				break;
			case 133: // "srm-timeout"
				if ((timeout_tmp = strtol (optarg, &dp, 10)) < 0 || 
						*dp != '\0' ) {
					fprintf (stderr,
							"invalid value for '--srm-timeout': %s\n", optarg);
					errflg++;
				} else
					gfal_set_timeout_srm (timeout_tmp);
				break;
			case 's':
				spacetokendesc = optarg;
				break;
			case 'D':
				if (!strcasecmp (optarg, "srmv1"))
					defaulttype = TYPE_SRM;
				else if (!strcasecmp (optarg, "srmv2"))
					defaulttype = TYPE_SRMv2;
				else if (!strcasecmp (optarg, "se"))
					defaulttype = TYPE_SE;
				else {
					fprintf (stderr,"Invalid SE type: %s\n", optarg);
				}
				break;
			case 'T':
				if (!strcmp (optarg, "srmv1"))
					setype = TYPE_SRM;
				else if (!strcmp (optarg, "srmv2"))
					setype = TYPE_SRMv2;
				else if (!strcmp (optarg, "se"))
					setype = TYPE_SE;
				else {
					fprintf (stderr,"Invalid SE type: %s\n", optarg);
					errflg++;
				}
				break;
			case 'b':
				nobdii = 1;
				break;
			case 'v':
				++verbose;
				break;
			case 'W':
				version = 1;
				break;
			case '?':
			default:
				errflg++;
				break;
		}
	}
	if (version) {
		printf ("lcg_util-%s\n", lcg_util_version ());
		printf ("GFAL-client-%s\n", gfal_version ());
		exit (0);
	}
	if (optind + 2 > argc)
		errflg++;
	if (errflg || helpflg) {
		fprintf (stderr, "usage: %s %s%s%s%s", argv[0],
				"[-h,--help] [-s,--st spacetokendesc] [-b,--nobdii] [--srm-timeout value]\n",    
				"\t[--connect-timeout value] [--sendreceive-timeout value] [--bdii-timeout value]\n",
				"\t[-D,--defaultsetype se|srmv1|srmv2] [-T,--setype se|srmv1|srmv2]\n",
				"\t[-v,--verbose] [--version] surl protocol [protocol [...]]\n");
		exit (errflg ? 1 : 0);
	}

	gfal_set_verbose (verbose);
    /* Terminate the argument list with NULL, just in case...*/
	argv[argc] = NULL;

	if (lcg_gt4 (argv[optind], defaulttype, setype, nobdii, argv + optind + 1, 
				&turl, &usedtype, &reqid, &fileid, &token, NULL, 0, timeout, spacetokendesc) < 0) {
		if (errno)
			perror ("lcg_gt");
		exit (1);
	}
	printf ("%s\n", turl);
	if (usedtype == TYPE_SRMv2) {
		if (token)
			printf ("%s\n", token);
		else
			printf ("no_request_token\n", token);
	} else if (usedtype == TYPE_SRM) {
		printf ("%d\n", reqid);
		printf ("%d\n", fileid);
	}
	exit (0);
}
