/*
 * Copyright (c) Members of the EGEE Collaboration. 2004.
 * See http://www.eu-egee.org/partners/ for details on the copyright holders.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *     http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#ifndef lint
static char sccsid[] = "@(#)$RCSfile: lcg-sd.c,v $ $Revision: 1.15 $ $Date: 2008/12/08 16:45:24 $ CERN Jean-Philippe Baud";
#endif /* not lint */

/*	lcg-sd - set file status to "Done" */
#include <sys/types.h>
#include <errno.h>
#include <getopt.h>
#include <stdlib.h>
#include <stdio.h>
#include "lcg_util.h"

	int
main(int argc, char **argv)
{
	int helpflg = 0;
	int insecure = 0;
	int verbose = 0;
	int version = 0;
	int c;
	int errflg = 0;
	static struct option longopts[] = {
		{"insecure", no_argument, NULL, 'i'},
		{"nobdii", no_argument, NULL, 'b'},
		{"timeout", required_argument, NULL, 't'},
		{"connect-timeout", required_argument, NULL, 130},
		{"sendreceive-timeout", required_argument, NULL, 131},
		{"bdii-timeout", required_argument, NULL, 132},
		{"srm-timeout", required_argument, NULL, 133},
		{"help", no_argument, NULL, 'h'},
		{"verbose", no_argument, NULL, 'v'},
		{"version", no_argument, NULL, 'W'},
		{0, 0, 0, 0}
	};
	int timeout = 0, timeout_tmp = 0;
	char *dp;
	int no_bdii_check = 0;

	while ((c = getopt_long (argc, argv, "ibt:hvW", longopts, NULL)) != EOF) {
		switch (c) {
			case 'i':
				insecure = 1;
				break;
			case 'b':
				no_bdii_check = 1;
				break;
			case 't':
				fprintf (stderr, "Warning: -t,--timeout is deprecated! Use --timeout-* options instead\n");
				if ((timeout = strtol (optarg, &dp, 10)) < 0 || 
						*dp != '\0' ) {
					fprintf (stderr,
							"invalid value for '--timeout': %s\n", optarg);
					errflg++;
				}
			case 130: // "connect-timeout"
				if ((timeout_tmp = strtol (optarg, &dp, 10)) < 0 || 
						*dp != '\0' ) {
					fprintf (stderr,
							"invalid value for '--connect-timeout': %s\n", optarg);
					errflg++;
				} else
					gfal_set_timeout_connect (timeout_tmp);
				break;
			case 131: // "sendreceive-timeout"
				if ((timeout_tmp = strtol (optarg, &dp, 10)) < 0 || 
						*dp != '\0' ) {
					fprintf (stderr,
							"invalid value for '--sendreceive-timeout': %s\n", optarg);
					errflg++;
				} else
					gfal_set_timeout_sendreceive (timeout_tmp);
				break;
			case 132: // "bdii-timeout"
				if ((timeout_tmp = strtol (optarg, &dp, 10)) < 0 || 
						*dp != '\0' ) {
					fprintf (stderr,
							"invalid value for '--bdii-timeout': %s\n", optarg);
					errflg++;
				} else
					gfal_set_timeout_bdii (timeout_tmp);
				break;
			case 133: // "srm-timeout"
				if ((timeout_tmp = strtol (optarg, &dp, 10)) < 0 || 
						*dp != '\0' ) {
					fprintf (stderr,
							"invalid value for '--srm-timeout': %s\n", optarg);
					errflg++;
				} else
					gfal_set_timeout_srm (timeout_tmp);
				break;
			case 'h':
				helpflg = 1;
				break;
			case 'v':
				++verbose;
				break;
			case 'W':
				version = 1;
				break;
			default:
				errflg++;
				break;
		}
	}
	if (version) {
		printf ("lcg_util-%s\n", lcg_util_version ());
		printf ("GFAL-client-%s\n", gfal_version ());
		exit (0);
	}
	if (optind + 4 != argc && optind + 3 != argc)
		errflg++;
	if (errflg || helpflg) {
		fprintf (stderr, "usage: %s %s%s%s\n", argv[0],
				"[--connect-timeout value] [--sendreceive-timeout value] [--bdii-timeout value]\n",
				"\t[--srm-timeout value] [-h,--help] [-v,--verbose] [-b,--nobdii]\n",
				"\t[--version] surl (reqid fileid | reqtoken) oflag");
		exit (errflg ? 1 : 0);
	}

	if (optind + 3 == argc) { // request token => SRMv2
		if (lcg_sd3 (argv[optind], no_bdii_check, 0, 0, 
					argv[optind + 1], NULL, 0, timeout) < 0) {
			if (errno)
				perror ("lcg_sd");
			exit (1);
		}
	} else { // optind + 4 == argc => request id + file id => SRMv1
		if (lcg_sd3 (argv[optind], no_bdii_check, atoi (argv[optind + 1]), atoi(argv[optind + 2]), 
					NULL, NULL, 0, timeout) < 0) {
			if (errno)
				perror ("lcg_sd");
			exit (1);
		}
	}
	exit (0);
}
