/*
 * Copyright (c) Members of the EGEE Collaboration. 2004.
 * See http://www.eu-egee.org/partners/ for details on the copyright holders.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#ifndef lint
static char sccsid[] = "@(#)$RCSfile: lcg_del.c,v $ $Revision: 1.53 $ $Date: 2009/07/08 15:40:05 $ CERN Jean-Philippe Baud";
#endif /* not lint */

/*	lcg_del - deletes one file (either one replica or all replicas).
 *	If it was the last or only valid replica, the entries corresponding to
 *	this GUID are also removed from the RMC.
 */
#include <errno.h>
#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include "gfal_api.h"
#include "lcg_util.h"

	int
se_match (char *pfn, char *se, char *errbuf, int errbufsz)
{
	int len;
	char *p;
	char server[256];

	if (strncmp (pfn, "srm:", 4) && strncmp (pfn, "sfn:", 4))
		return (1);
	if ((p = strchr (pfn + 6, '/')) == NULL) {
		gfal_errmsg (errbuf, errbufsz, GFAL_ERRLEVEL_ERROR, "Invalid Source URL: could not extract host name.");
		errno = EINVAL;
		return (-1);
	}
	if ((len = p - pfn - 6) >= sizeof(server)) {
		gfal_errmsg (errbuf, errbufsz, GFAL_ERRLEVEL_ERROR, "Host name too long.");
		errno = ENAMETOOLONG;
		return (-1);
	}
	strncpy (server, pfn + 6, len);
	server[len] = '\0';
	if ((p = strchr (server, ':'))) *p = '\0';
	if (strcmp (server, se))
		return (1);
	return (0);
}

lcg_del (char *file, int aflag, char *se, char *vo, char *conf_file,
		int insecure, int verbose)
{
	return (lcg_del4 (file, TYPE_NONE, TYPE_NONE, 0, 0, aflag, se, vo, conf_file,
				insecure, verbose, 0, NULL, 0));
}

lcg_delt (char *file, int aflag, char *se, char *vo, char *conf_file,
		int insecure, int verbose, int timeout)
{
	return (lcg_del4 (file, TYPE_NONE, TYPE_NONE, 0, 0, aflag, se, vo, conf_file,
				insecure, verbose, timeout, NULL, 0));
}

lcg_delx (char *file, int aflag, char *se, char *vo, char *conf_file,
		int insecure, int verbose, char *errbuf, int errbufsz)
{
	return (lcg_del4 (file, TYPE_NONE, TYPE_NONE, 0, 0, aflag, se, vo, conf_file,
				insecure, verbose, 0, errbuf,  errbufsz));
}

lcg_delxt (char *file, int aflag, char *se, char *vo, char *conf_file,
		int insecure, int verbose, int timeout, char *errbuf, int errbufsz)
{
	return (lcg_del4 (file, TYPE_NONE, TYPE_NONE, 0, 0, aflag, se, vo, conf_file,
				insecure, verbose, timeout, errbuf, errbufsz));
}

lcg_del3 (char *file, enum se_type defaulttype, enum se_type setype, int nobdii,
		int aflag, char *se, char *vo, char *conf_file, int insecure, int verbose,
		int timeout, char *errbuf, int errbufsz)
{
	return (lcg_del4 (file, defaulttype, setype, nobdii, 0, aflag, se, vo, conf_file,
				insecure, verbose, timeout, errbuf, errbufsz));
}

lcg_del4 (char *file, enum se_type defaulttype, enum se_type setype, int nobdii,
		int nolfc, int aflag, char *se, char *vo, char *conf_file, int insecure,
		int verbose, int timeout, char *errbuf, int errbufsz)
{
	lcg_filestatus *statuses = NULL;
	int rc = 0, sav_errno = 0;

	rc = lcg_del7 (1, &file, defaulttype, setype, nobdii, nolfc, aflag, se, vo,
			conf_file, insecure, verbose, 0, 0, timeout, &statuses, errbuf, errbufsz);

	if (statuses) {
		if (statuses[0].status != 0) {
			if (statuses[0].explanation) {
				gfal_errmsg (errbuf, errbufsz, GFAL_ERRLEVEL_ERROR, "%s: %s",
						statuses[0].file, statuses[0].explanation);
				free (statuses[0].explanation);
			} else {
				gfal_errmsg (errbuf, errbufsz, GFAL_ERRLEVEL_ERROR, "%s: %s",
						statuses[0].file, strerror (statuses[0].status));
			}

			if (statuses[0].file) free (statuses[0].file);
			sav_errno = statuses[0].status;
		}
		free (statuses);
	}

	errno = sav_errno;
	return (rc);
}

lcg_del5 (int nbfiles, char **files, enum se_type defaulttype, enum se_type setype, int nobdii,
		int nolfc, int aflag, char *se, char *vo, char *conf_file, int insecure,
		int verbose, int timeout, lcg_filestatus **finalstatuses, char *errbuf, int errbufsz)
{
	return (lcg_del7 (nbfiles, files, defaulttype, setype, nobdii, nolfc, aflag, se, vo,
			conf_file, insecure, verbose, 0, 0, timeout, finalstatuses, errbuf, errbufsz));
}

lcg_del6 (int nbfiles, char **files, enum se_type defaulttype, enum se_type setype, int nobdii,
		int nolfc, int aflag, char *se, char *vo, char *conf_file, int insecure, int verbose,
		int bool_force, int timeout, lcg_filestatus **finalstatuses, char *errbuf, int errbufsz)
{
	return (lcg_del7 (nbfiles, files, defaulttype, setype, nobdii, nolfc, aflag, se, vo,
			conf_file, insecure, verbose, bool_force, 0, timeout, finalstatuses, errbuf, errbufsz));
}

lcg_del7 (int nbfiles, char **files, enum se_type defaulttype, enum se_type setype, int nobdii,
		int nolfc, int aflag, char *se, char *vo, char *conf_file, int insecure, int verbose,
		int bool_force, int bool_rmdir, int timeout, lcg_filestatus **finalstatuses, char *errbuf, int errbufsz)
{
	int i, j, n, rc, nberrors = 0;
	int *results = NULL;
	gfal_file gfile = NULL;
	gfal_request req = NULL;
	gfal_internal gfal = NULL;
	gfal_filestatus *filestatuses = NULL;
	char errmsg[GFAL_ERRMSG_LEN];

	gfal_set_verbose (verbose);

	if (finalstatuses == NULL) {
		errno = EINVAL;
		return (-1);
	}
	*finalstatuses = NULL;

	if (nbfiles < 1 || files == NULL) {
		gfal_errmsg (errbuf, errbufsz, GFAL_ERRLEVEL_ERROR, "You must specify SURLs");
		errno = EINVAL;
		return (-1);
	}
	if (aflag && nolfc) {
		gfal_errmsg (errbuf, errbufsz, GFAL_ERRLEVEL_ERROR, "Options '-a' and '--nolfc' are incompatible");
		errno = EINVAL;
		return (-1);
	}
	if (bool_rmdir && (aflag || se)) {
		fprintf (stderr, "Options -d,--dir isn't compatible with '-a' and '-s,--se' options!");
		exit (1);
	}
	if (bool_rmdir && nbfiles > 1) {
		fprintf (stderr, "You can delete only ONE directory at a time");
		exit (1);
	}
	if (vo && vo[0]) {
		if (strlen (vo) > GFAL_VO_MAXLEN) {
			gfal_errmsg (errbuf, errbufsz, GFAL_ERRLEVEL_ERROR, "VO name too long.");
			errno = ENAMETOOLONG;
			return (-1);
		}
		if (gfal_set_vo (vo) < 0)
			return (-1);
	}

	/* Directories aren't registered in LFC */
	if (bool_rmdir) nolfc = 1;

	if (timeout > 0) {
		// for backward compatibility...
		gfal_set_timeout_connect (timeout);
		gfal_set_timeout_sendreceive (timeout);
		gfal_set_timeout_bdii (timeout);
		gfal_set_timeout_srm (timeout);
	}

	if (check_surls (nbfiles, files, errbuf, errbufsz) < 0)
		return (-1);

	gfal_set_nobdii (nobdii);
	gfal_errmsg (NULL, 0, GFAL_ERRLEVEL_WARN, "VO name: %s", gfal_get_vo (errmsg, GFAL_ERRMSG_LEN));

	if (strncmp (files[0], "srm:", 4) == 0 || strncmp (files[0], "sfn:", 4) == 0) {
		char **guids = NULL, **actual_surls = NULL;
		int *pos_finalstatuses;
		int nbsurls, pos_surls, pos_guids, pos_statuses;
		int *guid_statuses;

		if (aflag) {
			gfal_errmsg (errbuf, errbufsz, GFAL_ERRLEVEL_ERROR, "Option '-a' is invalid with SURLs");
			errno = EINVAL;
			return (-1);
		}
		if ((*finalstatuses = (lcg_filestatus *) calloc (nbfiles, sizeof (lcg_filestatus))) == NULL ||
				(pos_finalstatuses = (int *) calloc (nbfiles + 1, sizeof (int))) == NULL) {
			if (finalstatuses) {
				free (finalstatuses);
				*finalstatuses = NULL;
			}
			return (-1);
		}

		for (i = 0; i < nbfiles; ++i) {
			pos_finalstatuses[i] = i;
			(*finalstatuses)[i].file = strdup (files[i]);
		}

		if (!nolfc) {
			if ((actual_surls = (char **) calloc (nbfiles + 1, sizeof (char *))) == NULL) {
				if (actual_surls) free (actual_surls);
				return (-1);
			}

			if (gfal_guidsforpfns (nbfiles, (const char **) files, R_OK|W_OK, &guids, &guid_statuses, errbuf, errbufsz) < 0 ||
					guids == NULL || guid_statuses == NULL) {
				free (actual_surls);
				if (pos_finalstatuses) free (pos_finalstatuses);

				if (*finalstatuses) {
					for (i = 0; i < nbfiles; ++i) {
						if ((*finalstatuses)[i].file) free ((*finalstatuses)[i].file);
						if ((*finalstatuses)[i].explanation) free ((*finalstatuses)[i].explanation);
					}
					free (*finalstatuses);
					*finalstatuses = NULL;
				}

				return (-1);
			}

			for (pos_guids = 0, pos_surls = 0; pos_guids < nbfiles; ++pos_guids) {
				if (guids[pos_guids] != NULL) {
					actual_surls[pos_surls] = files[pos_guids];
					if (pos_surls < pos_guids) {
						guids[pos_surls] = guids[pos_guids];
						pos_finalstatuses[pos_surls] = pos_finalstatuses[pos_guids];
					}
					++pos_surls;
				} else {
					++nberrors;
					(*finalstatuses)[pos_finalstatuses[pos_guids]].status = guid_statuses[pos_guids];
				}
			}
			free (guid_statuses);
			nbsurls = pos_surls;
		} else
			nbsurls = nbfiles;

		if (nbsurls == 0) {
			if (actual_surls) free (actual_surls);
			if (pos_finalstatuses) free (pos_finalstatuses);
			return (!bool_force && nberrors ? -1 : 0);
		}

		if ((req = gfal_request_new ()) == NULL) {
			int i;

			if (actual_surls) free (actual_surls);
			if (pos_finalstatuses) free (pos_finalstatuses);
			if (guids) free (guids);

			if (*finalstatuses) {
				for (i = 0; i < nbfiles; ++i) {
					if ((*finalstatuses)[i].file) free ((*finalstatuses)[i].file);
					if ((*finalstatuses)[i].explanation) free ((*finalstatuses)[i].explanation);
				}
				free (*finalstatuses);
				*finalstatuses = NULL;
			}

			return (-1);
		}

		if (!nolfc) {
			req->nbfiles = nbsurls;
			req->surls = actual_surls;
		} else {
			req->nbfiles = nbfiles;
			req->surls = files;
		}
		req->defaultsetype = defaulttype;
		req->setype = setype;
		req->no_bdii_check = nobdii;

		if (gfal_init (req, &gfal, errbuf, errbufsz) < 0) {
			if (actual_surls) free (actual_surls);
			if (pos_finalstatuses) free (pos_finalstatuses);
			if (guids) free (guids);

			if (*finalstatuses) {
				for (i = 0; i < nbfiles; ++i) {
					if ((*finalstatuses)[i].file) free ((*finalstatuses)[i].file);
					if ((*finalstatuses)[i].explanation) free ((*finalstatuses)[i].explanation);
				}
				free (*finalstatuses);
				*finalstatuses = NULL;
			}

			free (req);
			return (-1);
		}
		free (req);
		gfal_errmsg (NULL, 0, GFAL_ERRLEVEL_WARN, "SE type: %s", lcg_se_type_string (gfal->setype));

		if (!bool_rmdir)
			rc = gfal_deletesurls (gfal, errbuf, errbufsz);
		else
			rc = gfal_removedir (gfal, errbuf, errbufsz);

		if (rc < 0 || (n = gfal_get_results (gfal, &filestatuses)) < 0 || filestatuses == NULL) {
			int i;

			if (actual_surls) free (actual_surls);
			if (pos_finalstatuses) free (pos_finalstatuses);
			if (guids) free (guids);
			gfal_internal_free (gfal);

			if (*finalstatuses) {
				for (i = 0; i < nbfiles; ++i) {
					if ((*finalstatuses)[i].file) free ((*finalstatuses)[i].file);
					if ((*finalstatuses)[i].explanation) free ((*finalstatuses)[i].explanation);
				}
				free (*finalstatuses);
				*finalstatuses = NULL;
			}

			errno = EFAULT;
			return (-1);
		}

		for (pos_surls = 0, i = 0; i < nbsurls; ++i) {
			if (!nolfc) {
				if (strcmp (filestatuses[i].surl, actual_surls[i]) == 0)
					pos_statuses = i;
				else {
					for (pos_statuses = 0; pos_statuses < nbsurls; ++pos_statuses) {
						if (strcmp (filestatuses[pos_statuses].surl, actual_surls[i]) == 0)
							break;
					}
					/* If we didn't find the position, it may mean that SURLs in results structure
					 * are a bit different, so let's assume that the order was unchanged */
					if (pos_statuses == nbsurls)
						pos_statuses = i;
				}
			} else
				pos_statuses = i;

			if (filestatuses[pos_statuses].status == 0) {
				if (!nolfc) {
					gfal_errmsg (NULL, 0, GFAL_ERRLEVEL_WARN, "[guid:%s] %s - DELETED", guids[i], filestatuses[pos_statuses].surl);
					if (pos_surls < i) {
						guids[pos_surls] = guids[i];
						actual_surls[pos_surls] = actual_surls[i];
						pos_finalstatuses[pos_surls] = pos_finalstatuses[i];
					}
					++pos_surls;
				} else
					gfal_errmsg (NULL, 0, GFAL_ERRLEVEL_WARN, "%s - DELETED", filestatuses[pos_statuses].surl);
			} else {
				++nberrors;
				if (!nolfc) {
					if (bool_force && filestatuses[pos_statuses].status == ENOENT) {
						if (pos_surls < i) {
							guids[pos_surls] = guids[i];
							actual_surls[pos_surls] = actual_surls[i];
							pos_finalstatuses[pos_surls] = pos_finalstatuses[i];
						}
						++pos_surls;
					} else {
						free (guids[i]);
					}
				}

				if (filestatuses[pos_statuses].explanation)
					snprintf (errmsg, GFAL_ERRMSG_LEN, "%s: %s", filestatuses[pos_statuses].surl, filestatuses[pos_statuses].explanation);
				else
					snprintf (errmsg, GFAL_ERRMSG_LEN, "%s: %s", filestatuses[pos_statuses].surl, strerror (filestatuses[pos_statuses].status));

				(*finalstatuses)[pos_finalstatuses[i]].status = filestatuses[pos_statuses].status;
				(*finalstatuses)[pos_finalstatuses[i]].explanation = strdup (errmsg);
			}
		}

		gfal_internal_free (gfal);
		nbsurls = pos_surls;
		rc = 0;

		if (!nolfc) {
			int sav_errno = 0;
			rc = gfal_unregister_pfns (nbsurls, (const char **) guids,
					(const char **) actual_surls, &results, errmsg, GFAL_ERRMSG_LEN);
			sav_errno = errno;

			for (pos_surls = 0; pos_surls < nbsurls; ++pos_surls) {
				free (guids[pos_surls]);
				if (rc == 0 && results != NULL) {
					(*finalstatuses)[pos_finalstatuses[i]].status = results[pos_surls];
				} else {
					(*finalstatuses)[pos_finalstatuses[i]].status = sav_errno;
					(*finalstatuses)[pos_finalstatuses[i]].explanation = errmsg[0] ? strdup (errmsg) : NULL;
				}
			}

			free (actual_surls);
			free (guids);
			if (results) free (results);
		}

		free (pos_finalstatuses);
		return (!bool_force && nberrors ? -1 : 0);
	}

	if ((strncmp (files[0], "lfn:", 4) != 0 && strncmp (files[0], "guid:", 5) != 0)) {
		gfal_errmsg (errbuf, errbufsz, GFAL_ERRLEVEL_ERROR, "File starts with neither \"lfn:\", \"guid:\", \"srm:\" nor \"sfn:\".");
		errno = EINVAL;
		return (-1);
	}
	if (nolfc) {
		gfal_errmsg (errbuf, errbufsz, GFAL_ERRLEVEL_ERROR, "With '--nolfc' option, you can only specify SURLs (srm: or sfn:)");
		errno = EINVAL;
		return (-1);
	}
	if (nobdii && !gfal_is_purifydisabled ()) {
		gfal_errmsg (errbuf, errbufsz, GFAL_ERRLEVEL_ERROR, "Option '--nobdii' is not compatible with lfn: and guid:");
		errno = EINVAL;
		return (-1);
	}
	if (aflag == 0 && se == NULL) {
		gfal_errmsg (errbuf, errbufsz, GFAL_ERRLEVEL_ERROR, "To delete LFNs or GUIDs, you must specify '-a' option");
		errno = EINVAL;
		return (-1);
	}

	if ((*finalstatuses = (lcg_filestatus *) calloc (nbfiles, sizeof (lcg_filestatus))) == NULL)
		return (-1);

	for (i = 0; i < nbfiles; ++i) {
		(*finalstatuses)[i].file = strdup (files[i]);
		if ((gfile = gfal_file_new (files[i], NULL, 0, errmsg, GFAL_ERRMSG_LEN)) == NULL) {
			++nberrors;
			(*finalstatuses)[i].explanation = errmsg[0] ? strdup (errmsg) : NULL;
			(*finalstatuses)[i].status = errno;
			continue;
		}
		if (gfile->errcode != 0) {
			++nberrors;
			if (gfile->errmsg)
				(*finalstatuses)[i].explanation = strdup (gfile->errmsg);
			(*finalstatuses)[i].status = gfile->errcode;
			gfal_file_free (gfile);
			continue;
		}
		if (gfal_access (files[i], R_OK|W_OK) < 0) {
			++nberrors;
			(*finalstatuses)[i].status = errno;
			continue;
		}

		if (req == NULL && gfile->nbreplicas > 0) {
			if ((req = gfal_request_new ()) == NULL) {
				if (*finalstatuses) {
					free (*finalstatuses);
					*finalstatuses = NULL;
				}
				gfal_file_free (gfile);
				return (-1);
			}
			req->nbfiles = 1;
			req->defaultsetype = defaulttype;
			req->setype = setype;
			req->no_bdii_check = nobdii;
		}

		if (gfile->nbreplicas == 0)
			if (unregister_alias (gfile->guid, gfile->lfn, errbuf, errbufsz) < 0)
				++nberrors;

		for (j = 0; j < gfile->nbreplicas; ++j) {
			if (gfile->replicas[j] == NULL || (!aflag && se_match (gfile->replicas[j]->surl, se, errbuf, errbufsz) != 0))
				continue;

			req->surls = &(gfile->replicas[j]->surl);

			if (gfal_init (req, &gfal, errbuf, errbufsz) < 0)
			{
                gfal_errmsg (NULL, 0, GFAL_ERRLEVEL_ERROR, "%s", errbuf);
				++nberrors;
				continue;
			}

			gfal_errmsg (NULL, 0, GFAL_ERRLEVEL_WARN, "SE type: %s", lcg_se_type_string (gfal->setype));

			if (gfal_deletesurls (gfal, errbuf, errbufsz) < 0) {
				++nberrors;
				gfal_internal_free (gfal);
				if (gfile->lfn)
					gfal_errmsg (NULL, 0, GFAL_ERRLEVEL_WARN, "[lfn:%s] %s - ERROR", gfile->lfn, gfile->replicas[j]->surl);
				else if (gfile->guid)
					gfal_errmsg (NULL, 0, GFAL_ERRLEVEL_WARN, "[guid:%s] %s - ERROR", gfile->guid, gfile->replicas[j]->surl);
				continue;
			}
			if ((n = gfal_get_results (gfal, &filestatuses)) < 0 || filestatuses == NULL) {
				++nberrors;
				gfal_internal_free (gfal);
				if (gfile->lfn)
					gfal_errmsg (errbuf, errbufsz, GFAL_ERRLEVEL_ERROR,
							"[lfn:%s] %s - Internal error", gfile->lfn, gfile->replicas[j]->surl);
				if (gfile->guid)
					gfal_errmsg (errbuf, errbufsz, GFAL_ERRLEVEL_ERROR,
							"[guid:%s] %s - Internal error", gfile->guid, gfile->replicas[j]->surl);
				continue;
			}
			if (filestatuses[0].status != 0) {
				++nberrors;
				if (gfile->lfn)
					gfal_errmsg (NULL, 0, GFAL_ERRLEVEL_WARN, "[lfn:%s] %s - %s", gfile->lfn, gfile->replicas[j]->surl,
							filestatuses[0].explanation ? filestatuses[0].explanation : strerror (filestatuses[0].status));
				else if (gfile->guid)
					gfal_errmsg (NULL, 0, GFAL_ERRLEVEL_WARN, "[guid:%s] %s - %s", gfile->guid, gfile->replicas[j]->surl,
							filestatuses[0].explanation ? filestatuses[0].explanation : strerror (filestatuses[0].status));

				if (!bool_force || filestatuses[0].status != ENOENT) {
					gfal_internal_free (gfal);
					continue;
				}
			} else if (gfile->lfn)
				gfal_errmsg (NULL, 0, GFAL_ERRLEVEL_WARN, "[lfn:%s] %s - DELETED", gfile->lfn, gfile->replicas[j]->surl);
			else if (gfile->guid)
				gfal_errmsg (NULL, 0, GFAL_ERRLEVEL_WARN, "[guid:%s] %s - DELETED", gfile->guid, gfile->replicas[j]->surl);

			gfal_internal_free (gfal);
			gfal_unregister_pfns (1, (const char **) &(gfile->guid), (const char **) &(gfile->replicas[j]->surl), &results, errbuf, errbufsz);
			if (results) free (results);
		}

		gfal_file_free (gfile);
	}

	if (req) free (req);
	return (!bool_force && nberrors ? -1 : 0);
}
