/*
 * Copyright (c) Members of the EGEE Collaboration. 2004.
 * See http://www.eu-egee.org/partners/ for details on the copyright holders.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *     http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
 

#ifndef lint
static char sccsid[] = "@(#)$RCSfile: lcg_globus_ifce.c,v $ $Revision: 1.35 $ $Date: 2009/08/27 09:55:20 $ CERN Jean-Philippe Baud";
#endif /* not lint */

/*      lcg_globus_ifce - wrapper for globus APIs */
#include <errno.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include "gfal_api.h"
#include "lcg_util.h"
#include "gridftp-ifce.h"



static int lcg_url_is_for_globus(const char * url){
    const char* gsiftp_proto = "gsiftp://";
    const char* file_proto = "file:";	

	return ((strncmp(url, gsiftp_proto, strlen(gsiftp_proto)) == 0)
			|| (strncmp(url, file_proto, strlen(file_proto)) == 0));

}

 

int
getfilesize(char *file, GFAL_LONG64 *size, char *errbuf, int errbufsz) {

    return (getfilesizet(file, size, errbuf, errbufsz, 0));

}



int
getfilesizet(char *file, GFAL_LONG64 *size, char *errbuf, int errbufsz, int timeout) {
    do_globus_activates();
    globus_result_t result =  dmutils_gridftp_get_filesize(
            file,
            (globus_off_t*) size,
            errbuf,
            errbufsz,
            timeout);
     do_globus_deactivates();
     return result == GLOBUS_TRUE ? (0) : (-1);
}

int
makedir(char *dir, char *errbuf, int errbufsz) {
    return (makedirt(dir, errbuf, errbufsz, 0));

}

int
makedirt(char *dir, char *errbuf, int errbufsz, int timeout) {
    do_globus_activates();
    globus_result_t result = dmutils_gridftp_make_dir(
            dir,
            errbuf,
            errbufsz,
            timeout);
      do_globus_deactivates();   
 
     return result == GLOBUS_TRUE ? (0) : (-1);
}

int
makedirp(char *dest_file, char *errbuf, int errbufsz) {
    return (makedirpt(dest_file, errbuf, errbufsz, 0));

}

int
makedirpt(char *dest_file, char *errbuf, int errbufsz, int timeout) {
    int c = 0;
    char *lastslash = NULL;
    char *p;
    char *q;
    char *localbuf;
    int localbufsz;
    int save_errno = 0;

    if (errbuf != NULL) {
        localbufsz = errbufsz;
    } else {
        localbufsz = 8192;
    }
    localbuf = (char *) malloc(localbufsz);
    if (localbuf == NULL)
        return (-1);
    localbuf[0] = '\0';

    p = strrchr(dest_file, '/');
    while (p) {
        if (lastslash == NULL) lastslash = p;
        if (*(p - 1) == ':') break;
        *p = '\0';
        c = makedirt(dest_file, localbuf, localbufsz, timeout);
        if (c && errno == EEXIST) c = 0;
        if (c == 0) {
            *p = '/';
            break;
        }
        if (errno != ENOENT) {
            *p = '/';
            goto ret;
        }
        q = strrchr(dest_file, '/');
        *p = '/';
        p = q;
    }
    while (c == 0 && p < lastslash && (p = strchr(p + 1, '/'))) {
        *p = '\0';
        c = makedirt(dest_file, localbuf, localbufsz, timeout);
        if (c && errno == EEXIST) c = 0;
        *p = '/';
    }
ret:
    save_errno = errno;
    if (c && strlen(localbuf) > 0)
        gfal_errmsg(errbuf, errbufsz, GFAL_ERRLEVEL_ERROR, "%s", localbuf);

    free(localbuf);
    errno = save_errno;

    return (c);
}

int
copyfile(char *src_file, char *dest_file, int nbstreams, char *errbuf, int errbufsz) {

    return (copyfilex(src_file, dest_file, nbstreams, errbuf, errbufsz, 0, 0));
}


int
copyfilex_internal(char *src_file, char *dest_file, int nbstreams, char *errbuf,
        int errbufsz, int timeout, int verbose) {

    char* temp = 0;
    struct stat64 stat_buf;
    /*nested function to return dest file path (without protocol)*/
    char* substring(const char* str, size_t begin, size_t len) {
        if (str == 0 || strlen(str) == 0 || strlen(str) < begin || strlen(str) < (begin + len))
            return 0;

        return strndup(str + begin, len);
    }

    do_globus_activates();
    globus_result_t result = dmutils_gridftp_copy(
            src_file,
            dest_file,
            errbuf,
            errbufsz,
            timeout,
            nbstreams,
            0,
            0,
            verbose,
            NULL, NULL, NULL, NULL, NULL, NULL, NULL);
    do_globus_deactivates();
    /*transfer failed, check dest protocol(file:), and if dest file exists and it's zero size*/
    if ((GLOBUS_TRUE != result) && (strncmp(dest_file, "file:", 5) == 0)) {
        temp = substring(dest_file, 5, strlen(dest_file) - 5); /*remove protocol*/
        if (0 == stat64(temp, &stat_buf)) /*dest file shouldn't be here, delete it*/
            unlink(temp);

        free(temp);
    }
    return result == GLOBUS_TRUE ? (0) : (-1);
}


int lcg_copyfile_gfal_open(const char* src, int flag, char *errbuf, int errbufsz){
	int fd;
	errno =0;
	if( ( fd = gfal_open(src, flag, 0755)) <0){
		gfal_errmsg (errbuf, errbufsz, GFAL_ERRLEVEL_ERROR, "GFAL1: Impossible to Open %s: %s", src, strerror(errno));	
		fd = -1;
	}
	return fd;
}

int
copyfilex_internal_gfal(char *src_file, char *dest_file, int nbstreams, char *errbuf,
        int errbufsz, int timeout, int verbose) {			
	gfal_errmsg (NULL, 0, GFAL_ERRLEVEL_WARN, "Local copy triggered from %s to %s, use gfal 1.0", src_file, dest_file);
	int fd_src=-1, fd_dst=-1, ret = -1;
	if( ( (fd_src= lcg_copyfile_gfal_open(src_file, O_RDONLY, errbuf, errbufsz)) > 0)
	    && ( (fd_dst= lcg_copyfile_gfal_open(dest_file, O_WRONLY | O_CREAT, errbuf, errbufsz)) > 0) ){
		gfal_errmsg (NULL, 0, GFAL_ERRLEVEL_WARN, "Start Local transfer.....");
		gfal_set_verbose(0xff);
		
		const size_t buffer_size= 1 << 20;
		char buffer_gfal_copy[buffer_size];
		while(1){
			ssize_t s_read;
			errno = 0;
			s_read= gfal_read(fd_src, buffer_gfal_copy, buffer_size);
			if(s_read == 0){
				ret = 0;
				break;
			}
			if(s_read < 0){
				gfal_errmsg (errbuf, errbufsz, GFAL_ERRLEVEL_ERROR, "GFAL1: Read Error on %s: %s", src_file, strerror(errno));	
				break;	
			}
			
			// write 
			errno = 0;
			s_read= gfal_write(fd_dst, buffer_gfal_copy, s_read);
			if(s_read <= 0){
				gfal_errmsg (errbuf, errbufsz, GFAL_ERRLEVEL_ERROR, "GFAL1: Write Error on %s: %s", dest_file, strerror(errno));	
				break;	
			}		
		}
		

	}
	gfal_close(fd_src);	
	gfal_close(fd_dst);					
	return ret;
}


int
copyfilex(char *src_file, char *dest_file, int nbstreams, char *errbuf,
        int errbufsz, int timeout, int verbose) {

   if(lcg_url_is_for_globus(src_file) && lcg_url_is_for_globus(dest_file)){
	   return copyfilex_internal(src_file, dest_file, nbstreams, errbuf, errbufsz, timeout, verbose);
   }else{
	   return copyfilex_internal_gfal(src_file, dest_file, nbstreams, errbuf, errbufsz, timeout, verbose);
   }
}


int
gridftp_checksum(char *file, enum gfal_cksm_type cksmtype, char **cksm,
        int keep_conn, char *errbuf, int errbufsz, int timeout) {
    char temp_cksm[256] = {0};

    do_globus_activates();
    globus_result_t gresult = dmutils_gridftp_checksum(
            file,
            temp_cksm,
            gfal_get_checksum_algname(cksmtype),
            errbuf,
            errbufsz,
            timeout);
     do_globus_deactivates();

    if (gresult == GLOBUS_TRUE) {
        *cksm = strdup(temp_cksm);
        return (0);
    }

    return (-1);
}












