/*
 * Copyright (c) Members of the EGEE Collaboration. 2004.
 * See http://www.eu-egee.org/partners/ for details on the copyright holders.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *     http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#ifndef lint
static char sccsid[] = "@(#)$RCSfile: lcg_la.c,v $ $Revision: 1.16 $ $Date: 2008/11/10 12:38:16 $ CERN Jean-Philippe Baud";
#endif /* not lint */

/*	lcg_la - get the list of aliases for a given LFN, GUID or SURL */
#include <errno.h>
#include <stdlib.h>
#include <string.h>
#include "gfal_api.h"
#include "lcg_util.h"

lcg_la (char *file, char *vo, char *conf_file, int insecure, char ***lfns)
{
	return (lcg_la2 (file, 0, vo, lfns, NULL, 0));
}

lcg_lax (char *file, char *vo, char *conf_file, int insecure, char ***lfns,
	char *errbuf, int errbufsz)
{
	return (lcg_la2 (file, 0, vo, lfns, errbuf, errbufsz));
}

lcg_la2 (char *file, int nobdii, char *vo, char ***lfns, char *errbuf, int errbufsz)
{
	char *actual_lfn = NULL, *actual_guid = NULL;
	char actual_file[GFAL_PATH_MAXLEN];

	if (file == NULL || lfns == NULL) {
		errno = EFAULT;
		return (-1);
	}
	if (vo && vo[0]) {
		if (strlen (vo) > GFAL_VO_MAXLEN) {
			gfal_errmsg (errbuf, errbufsz, GFAL_ERRLEVEL_ERROR, "VO name too long.");
			errno = ENAMETOOLONG;
			return (-1);
		}
		if (gfal_set_vo (vo) < 0)
			return (-1);
	}

	if (canonical_url (file, NULL, actual_file, GFAL_PATH_MAXLEN, errbuf, GFAL_ERRMSG_LEN) < 0)
		return (-1);

	gfal_set_nobdii (nobdii);

	if (strncmp (actual_file, "lfn:", 4) == 0) {
		actual_lfn = strdup (actual_file);
	} else if (strncmp (actual_file, "guid:", 5) == 0) {
		actual_guid = strdup (actual_file);
	} else if (strncmp (actual_file, "srm:", 4) == 0 || strncmp (actual_file, "sfn:", 4) == 0) {
		if ((actual_guid = gfal_guidforpfn (actual_file, errbuf, errbufsz)) == NULL)
			return (-1);
	} else {
		gfal_errmsg (errbuf, errbufsz, GFAL_ERRLEVEL_ERROR,
				"%s: File starts with neither \"lfn:\", \"guid:\", \"srm:\" nor \"sfn:\".", file);
		errno = EINVAL;
		return (-1);
	}

	*lfns = NULL;
	*lfns = gfal_get_aliases (actual_lfn, actual_guid, errbuf, errbufsz);

	if (actual_lfn) free (actual_lfn);
	if (actual_guid) free (actual_guid);
	return (*lfns ? 0 : -1);
}
