/*
 * Copyright (c) Members of the EGEE Collaboration. 2004.
 * See http://www.eu-egee.org/partners/ for details on the copyright holders.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#ifndef lint
static char sccsid[] = "@(#)$RCSfile: lcg_rep.c,v $ $Revision: 1.58 $ $Date: 2009/12/14 15:31:30 $ CERN Jean-Philippe Baud";
#endif /* not lint */

/*	lcg_rep - copy a file from one Storage Element to another Storage
 *	Element and registers it in the LRC
 */
#include <errno.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <time.h>
#include <sys/stat.h>
#include <sys/times.h>
#include <unistd.h>
#include <uuid/uuid.h>
#include "gfal_api.h"
#include "lcg_util.h"
#include "gfal_dependencies.h"
#include "globus_dependencies.h"

lcg_rep (char *src_file, char *dest_file, char *vo, char *relative_path,
	int nbstreams, char *conf_file, int insecure, int verbose)
{
	return (lcg_rep4 (src_file, dest_file, TYPE_NONE, TYPE_NONE, TYPE_NONE, 0, vo, relative_path, nbstreams,
		conf_file, insecure, verbose, 0, NULL, NULL,  NULL, GFAL_CKSM_NONE, NULL, 0));
}

lcg_rept (char *src_file, char *dest_file, char *vo, char *relative_path,
	int nbstreams, char *conf_file, int insecure, int verbose, int timeout)
{
	return (lcg_rep4 (src_file, dest_file, TYPE_NONE, TYPE_NONE, TYPE_NONE, 0, vo, relative_path, nbstreams,
		conf_file, insecure, verbose, timeout, NULL, NULL, NULL, GFAL_CKSM_NONE, NULL, 0));
}

lcg_repx (char *src_file, char *dest_file, char *vo, char *relative_path,
	int nbstreams, char *conf_file, int insecure, int verbose, char *errbuf,
	int errbufsz)
{
	return (lcg_rep4 (src_file, dest_file, TYPE_NONE, TYPE_NONE, TYPE_NONE, 0, vo, relative_path, nbstreams,
		conf_file, insecure, verbose, 0, NULL, NULL, NULL, GFAL_CKSM_NONE, errbuf, errbufsz));

}

lcg_repxt (char *src_file, char *dest_file, char *vo, char *relative_path, int nbstreams,
	   char *conf_file, int insecure, int verbose, int timeout, char *errbuf, int errbufsz)
{
	return (lcg_rep4 (src_file, dest_file, TYPE_NONE, TYPE_NONE, TYPE_NONE, 0, vo, relative_path, nbstreams,
		conf_file, insecure, verbose, timeout, NULL, NULL, NULL, GFAL_CKSM_NONE, errbuf, errbufsz));
}

lcg_rep2 (char *src_file, char *dest_file, char *vo, char *relative_path, int nbstreams, char *conf_file, int insecure, int verbose,
	   int timeout, char *source_spacetokendesc, char *dest_spacetokendesc, char *errbuf, int errbufsz)
{
	return (lcg_rep4 (src_file, dest_file, TYPE_NONE, TYPE_NONE, TYPE_NONE, 0, vo, relative_path, nbstreams,
		conf_file, insecure, verbose, timeout, source_spacetokendesc, dest_spacetokendesc, NULL, GFAL_CKSM_NONE, errbuf, errbufsz));
}

lcg_rep3 (char *src_file, char *dest_file, enum se_type defaulttype, enum se_type srctype, enum se_type dsttype, int nobdii,
		char *vo, char *relative_path, int nbstreams, char *conf_file, int insecure, int verbose, int timeout,
		char *src_spacetokendesc, char *dest_spacetokendesc, char **actual_surl, char *errbuf, int errbufsz)
{
	return (lcg_rep4 (src_file, dest_file, defaulttype, srctype, dsttype, nobdii, vo, relative_path, nbstreams,
		conf_file, insecure, verbose, timeout, src_spacetokendesc, dest_spacetokendesc, actual_surl,
		GFAL_CKSM_NONE, errbuf, errbufsz));
}

int lcg_rep4 (char *src_file, char *dest_file, enum se_type defaulttype,
    enum se_type srctype, enum se_type dsttype, int nobdii, char *vo,
    char *relative_path, int nbstreams, char *conf_file, int insecure,
    int verbose, int timeout, char *src_spacetokendesc,
    char *dest_spacetokendesc, char **actual_surl, enum gfal_cksm_type cksmtype,
    char *errbuf, int errbufsz)
{
    return lcg_rep5 (src_file, dest_file, defaulttype, srctype, dsttype, nobdii,
        vo, relative_path, nbstreams, conf_file, insecure, verbose, timeout,
        src_spacetokendesc, dest_spacetokendesc, actual_surl, cksmtype,
        NULL, NULL, errbuf, errbufsz);
}

lcg_rep5 (char *src_file, char *dest_file, enum se_type defaulttype,
    enum se_type srctype, enum se_type dsttype, int nobdii, char *vo,
    char *relative_path, int nbstreams, char *conf_file, int insecure,
    int verbose, int timeout, char *src_spacetokendesc,
    char *dest_spacetokendesc, char **actual_surl, enum gfal_cksm_type cksmtype,
    char** src_protocols, char** dst_protocols, char *errbuf, int errbufsz)
{
	char dest_turl[GFAL_PATH_MAXLEN];
	char dest_surl_cat[GFAL_PATH_MAXLEN];
	off64_t filesize = 0;
	int i, n, sav_errno = 0;
	static char *proto4copy[] = {"gsiftp", NULL};
	clock_t tm_start = 0;
	struct tms tmp_tms;
	gfal_request src_req = NULL, dst_req = NULL;
	gfal_internal dst_gfal = NULL;
	gfal_filestatus *filestatuses;
	char errmsg[GFAL_ERRMSG_LEN];
	gfal_file gfile_src;
	int bool_issurlok = 0;
	char *current_surl = NULL;
	enum se_type setype = 0;
	char *srmv2_reqtoken= NULL;
	int srmv1_reqid = 0;
	char *src_cksm = NULL, *dst_cksm = NULL;
	enum gfal_cksm_type tmp_cksmtype = GFAL_CKSM_NONE;

	dest_turl[0] = 0;
	dest_surl_cat[0] = 0;

	if (!src_protocols)
        src_protocols = proto4copy;

    if (!dst_protocols)
        dst_protocols = proto4copy;

	gfal_set_verbose (verbose);

	if (src_file == NULL || dest_file == NULL) {
		errno = EFAULT;
		return (-1);
	}
	if (vo && vo[0]) {
		if (strlen (vo) > GFAL_VO_MAXLEN) {
			gfal_errmsg (errbuf, errbufsz, GFAL_ERRLEVEL_ERROR, "VO name too long.");
			errno = ENAMETOOLONG;
			return (-1);
		}
		if (gfal_set_vo (vo) < 0)
			return (-1);
	}

	if (timeout > 0) {
		// for backward compatibility...
		gfal_set_timeout_connect (timeout);
		gfal_set_timeout_sendreceive (timeout);
		gfal_set_timeout_bdii (timeout);
		gfal_set_timeout_srm (timeout);
	}

	if (nbstreams < 1) {
		gfal_errmsg (errbuf, errbufsz, GFAL_ERRLEVEL_ERROR, "Number of streams is < 1.");
		errno = EINVAL;
		return (-1);
	}

	gfal_set_nobdii (nobdii);

	/* check for dest file, if not there */
	if (dest_file == NULL) {
		errno = 0;
		if ((dest_file = get_default_se (errbuf, errbufsz)) == NULL) {
			if (errno == 0)
				gfal_errmsg (errbuf, errbufsz, GFAL_ERRLEVEL_ERROR,
						"No destination SE specified, and no default SE specified in environment\n");
			return (-1);
		}
	}

	/* get source GUID and source SURL */

	if ((gfile_src = lcg_gfal_functions.call_gfal_file_new (src_file, NULL, 0, errbuf, errbufsz)) == NULL)
		return (-1);
	if (gfile_src->errcode != 0) {
		sav_errno = gfile_src->errcode;
		gfal_file_free (gfile_src);
		errno = sav_errno;
		return (-1);
	}

	gfal_errmsg (NULL, 0, GFAL_ERRLEVEL_WARN, "Using grid catalog type: %s", gfal_file_get_catalog_name (gfile_src));
	gfal_errmsg (NULL, 0, GFAL_ERRLEVEL_WARN, "Using grid catalog : %s", get_catalog_endpoint (errmsg, GFAL_ERRMSG_LEN));
	gfal_errmsg (NULL, 0, GFAL_ERRLEVEL_WARN, "VO name: %s", gfal_get_vo (errmsg, GFAL_ERRMSG_LEN));
	gfal_errmsg (NULL, 0, GFAL_ERRLEVEL_WARN, "Checksum type: %s", gfal_get_checksum_algname (cksmtype));

	if (!gfile_src->guid && gfile_src->nbreplicas == 0) {
		gfal_errmsg (errbuf, errbufsz, GFAL_ERRLEVEL_ERROR, "%s: Invalid URL, must be a LFN, a GUID or a SURL");
		gfal_file_free (gfile_src);
		errno = EINVAL;
		return (-1);
	}
	if (!gfile_src->guid &&
			(gfile_src->guid = lcg_gfal_functions.call_gfal_guidforpfn (gfile_src->replicas[0]->surl, errbuf, errbufsz)) == NULL) {
		gfal_file_free (gfile_src);
		return (-1);
	}

	/* check if the destination replica already exists */

	if (gfile_src->nbreplicas > 0) {
		char *desthost = NULL;
		int found = 0;

		if (strstr (dest_file, ":/") == NULL)
			desthost = strdup (dest_file);
		else
			desthost = gfal_get_hostname (dest_file, errbuf, errbufsz);
		if (desthost == NULL) {
			sav_errno = errno;
			gfal_file_free (gfile_src);
			errno = sav_errno;
			return (-1);
		}

		for (i = 0; i < gfile_src->nbreplicas && gfile_src->replicas[i]; ++i) {
			if (!found) {
				if (se_match (gfile_src->replicas[i]->surl, desthost, errbuf, errbufsz) == 0)
					found = 1;
			}
		}
		free (desthost);

		if (found) {
			gfal_file_free (gfile_src);
			gfal_errmsg (NULL, 0, GFAL_ERRLEVEL_WARN, "Replica exists already on that target host");
			return (0);
		}
	}

	/* get source TURL */

	if (gfile_src->turl == NULL) {
		if ((src_req = gfal_request_new ()) == NULL) {
			sav_errno = gfile_src->errcode;
			gfal_file_free (gfile_src);
			errno = sav_errno;
			return (-1);
		}
		src_req->nbfiles = 1;
		src_req->defaultsetype = defaulttype;
		src_req->setype = srctype;
		src_req->no_bdii_check = nobdii;
		src_req->protocols = src_protocols;
		src_req->srmv2_spacetokendesc = src_spacetokendesc;
	}

	while (!bool_issurlok && gfile_src->errcode == 0) {
		bool_issurlok = 1;

		if (gfile_src->nbreplicas > 0) {
			if (gfile_src->gobj == NULL) {
				current_surl = (char *) gfal_file_get_replica (gfile_src);
				src_req->surls = (char **) &current_surl;
				gfal_errmsg (NULL, 0, GFAL_ERRLEVEL_WARN, "Trying SURL %s ...", current_surl);

				*errmsg = 0;
				if (!(bool_issurlok = gfal_init (src_req, &gfile_src->gobj, errmsg, GFAL_ERRMSG_LEN) >= 0))
					gfal_file_set_replica_error (gfile_src, errno, errmsg);
			}

			*errmsg = 0;
			if (bool_issurlok && !(bool_issurlok = lcg_gfal_functions.call_gfal_ls (gfile_src->gobj, errmsg, GFAL_ERRMSG_LEN) >= 0))
				gfal_file_set_replica_error (gfile_src, errno, errmsg);

			if (bool_issurlok && (!(bool_issurlok = lcg_gfal_functions.call_gfal_get_results (gfile_src->gobj, &filestatuses) >= 0) ||
						!(bool_issurlok = filestatuses != NULL)))
				gfal_file_set_replica_error (gfile_src, errno, "Internal error");

			if (bool_issurlok && !(bool_issurlok = filestatuses[0].status == 0))
				gfal_file_set_replica_error (gfile_src, filestatuses[0].status, filestatuses[0].explanation);

			if (bool_issurlok) {
				filesize = filestatuses[0].stat.st_size;
				if (cksmtype != GFAL_CKSM_NONE) {
					if ((tmp_cksmtype = gfal_parse_checksum_algname (filestatuses[0].checksumtype)) == cksmtype) {
						/* checksum returned by srmLs is of the same type as the requested one... */
						src_cksm = strdup (filestatuses[0].checksum);
					}
				}
			}

			*errmsg = 0;
			if (bool_issurlok && !(bool_issurlok = lcg_gfal_functions.call_gfal_turlsfromsurls (gfile_src->gobj, errmsg, GFAL_ERRMSG_LEN) >= 0))
				gfal_file_set_replica_error (gfile_src, errno, errmsg);

			if (bool_issurlok && (!(bool_issurlok = lcg_gfal_functions.call_gfal_get_results (gfile_src->gobj, &filestatuses) >= 0) ||
						!(bool_issurlok = filestatuses != NULL)))
				gfal_file_set_replica_error (gfile_src, errno, "Internal error");

			if (bool_issurlok && !(bool_issurlok = filestatuses[0].status == 0))
				gfal_file_set_replica_error (gfile_src, filestatuses[0].status, filestatuses[0].explanation);

			if (bool_issurlok) {
				gfal_errmsg (NULL, 0, GFAL_ERRLEVEL_WARN, "Source SE type: %s", lcg_se_type_string (gfile_src->gobj->setype));

				if (gfal_get_ids_setype (gfile_src->gobj, &setype, &srmv1_reqid, NULL, &srmv2_reqtoken) < 0)
					gfal_errmsg (NULL, 0, GFAL_ERRLEVEL_WARN, "Source SRM Request Token: [ERROR]");
				else if (setype == TYPE_SRM)
					gfal_errmsg (NULL, 0, GFAL_ERRLEVEL_WARN, "Source SRM Request Token: %d", srmv1_reqid);
				else if (setype == TYPE_SRMv2)
					gfal_errmsg (NULL, 0, GFAL_ERRLEVEL_WARN, "Source SRM Request Token: %s", srmv2_reqtoken ? srmv2_reqtoken: "(null)");

				if (srmv2_reqtoken) free (srmv2_reqtoken);

				gfile_src->turl = strdup (filestatuses[0].turl);
				if (!(bool_issurlok = gfile_src->turl != NULL)) {
					sav_errno = errno;
					if (dst_gfal) {
						gfal_abortrequest (dst_gfal, errmsg, GFAL_ERRMSG_LEN);
						gfal_deletesurls (dst_gfal, errmsg, GFAL_ERRMSG_LEN);
						gfal_internal_free (dst_gfal);
					}
					gfal_file_free (gfile_src);
					free (src_req);
					if (src_cksm) free (src_cksm);
					errno = sav_errno;
					return (-1);
				}
			}

			*errmsg = 0;
			if (bool_issurlok && !(bool_issurlok = gfal_set_xfer_running (gfile_src->gobj, errmsg, GFAL_ERRMSG_LEN) >= 0))
				gfal_file_set_replica_error (gfile_src, errno, errmsg);

			if (bool_issurlok && (!(bool_issurlok = lcg_gfal_functions.call_gfal_get_results (gfile_src->gobj, &filestatuses) >= 0) ||
						!(bool_issurlok = filestatuses != NULL)))
				gfal_file_set_replica_error (gfile_src, errno, "Internal error");

			if (bool_issurlok && !(bool_issurlok = filestatuses[0].status == 0))
				gfal_file_set_replica_error (gfile_src, filestatuses[0].status, filestatuses[0].explanation);
		} else if (gfile_src->turl != NULL) {
			struct stat64 statbuf;

			/* Get file size */
			if (!strncmp (gfile_src->turl, "gsiftp:", 7)) {
				*errmsg = 0;
				if (!(bool_issurlok = getfilesizet (gfile_src->turl, &filesize, errmsg, GFAL_ERRMSG_LEN, gfal_get_timeout_sendreceive ()) == 0))
					gfal_file_set_turl_error (gfile_src, errno, errmsg);
			} else {
				if (!(bool_issurlok = gfal_access (gfile_src->turl, R_OK) == 0))
					gfal_file_set_turl_error (gfile_src, errno, NULL);

				if (bool_issurlok && !(bool_issurlok = gfal_stat64 (gfile_src->turl, &statbuf) == 0))
					gfal_file_set_turl_error (gfile_src, errno, NULL);

				if (bool_issurlok && !(bool_issurlok = !S_ISDIR(statbuf.st_mode)))
					gfal_file_set_turl_error (gfile_src, EISDIR, NULL);

				filesize = statbuf.st_size;
			}
		} else bool_issurlok = 0;

		if (dest_turl[0] == '\0') {

			/* get destination TURL */

			if ((dst_req = gfal_request_new ()) == NULL)
				return (-1);
			dst_req->relative_path = relative_path;
			dst_req->nbfiles = 1;
			dst_req->oflag = 1;
			dst_req->filesizes = &filesize;
			dst_req->defaultsetype = defaulttype;
			dst_req->setype = dsttype;
			dst_req->no_bdii_check = nobdii;
			dst_req->protocols = dst_protocols;
			dst_req->srmv2_spacetokendesc = dest_spacetokendesc;

			if (strstr (dest_file, ":/") == NULL) { /* destination is just a hostname, one must generate a SURL */
				dst_req->endpoint = dest_file;
				dst_req->generatesurls = 1;
			} else {
				if (strncmp (dest_file, "srm:/", 5) &&
						strncmp (dest_file, "sfn:/", 5)) {
					gfal_errmsg (errbuf, errbufsz, GFAL_ERRLEVEL_ERROR, "Destination URL starts with neither 'srm:/' nor 'sfn:/'.");
					free (dst_req);
					free (src_req);
					gfal_file_free (gfile_src);
					if (src_cksm) free (src_cksm);
					errno = EINVAL;
					return (-1);
				}
				dst_req->surls = (char **) &dest_file;
			}

			if (gfal_init (dst_req, &dst_gfal, errbuf, errbufsz) < 0) {
				sav_errno = errno;
				gfal_file_free (gfile_src);
				free (dst_req);
				free (src_req);
				if (src_cksm) free (src_cksm);
				errno = sav_errno;
				return (-1);
			}
			free (dst_req);

			gfal_errmsg (NULL, 0, GFAL_ERRLEVEL_WARN, "Destination SE type: %s", lcg_se_type_string (dst_gfal->setype));

			if (purify_surl (dst_gfal->surls[0], dest_surl_cat, GFAL_PATH_MAXLEN) < 0) {
				sav_errno = errno;
				free (src_req);
				gfal_file_free (gfile_src);
				gfal_internal_free (dst_gfal);
				if (src_cksm) free (src_cksm);
				errno = sav_errno;
				return (-1);
			}

			if (actual_surl && (*actual_surl = strdup (dest_surl_cat)) == NULL) {
				free (src_req);
				gfal_file_free (gfile_src);
				gfal_internal_free (dst_gfal);
				if (src_cksm) free (src_cksm);
				errno = ENOMEM;
				return (-1);
			}

			if (lcg_gfal_functions.call_gfal_turlsfromsurls (dst_gfal, errbuf, errbufsz) < 0) {
				sav_errno = errno;
				free (src_req);
				gfal_file_free (gfile_src);
				gfal_internal_free (dst_gfal);
				if (src_cksm) free (src_cksm);
				errno = sav_errno;
				return (-1);
			}

			if (gfal_get_ids_setype (dst_gfal, &setype, &srmv1_reqid, NULL, &srmv2_reqtoken) < 0)
				gfal_errmsg (NULL, 0, GFAL_ERRLEVEL_WARN, "Destination SRM Request Token: [ERROR]");
			else if (setype == TYPE_SRM)
				gfal_errmsg (NULL, 0, GFAL_ERRLEVEL_WARN, "Destination SRM Request Token: %d", srmv1_reqid);
			else if (setype == TYPE_SRMv2)
				gfal_errmsg (NULL, 0, GFAL_ERRLEVEL_WARN, "Destination SRM Request Token: %s", srmv2_reqtoken ? srmv2_reqtoken: "(null)");

			if (srmv2_reqtoken) free (srmv2_reqtoken);

			if ((n = lcg_gfal_functions.call_gfal_get_results (dst_gfal, &filestatuses)) < 0 || filestatuses == NULL) {
				sav_errno = errno;
				free (src_req);
				gfal_file_free (gfile_src);
				gfal_internal_free (dst_gfal);
				gfal_errmsg (errbuf, errbufsz, GFAL_ERRLEVEL_ERROR, "Internal error");
				if (src_cksm) free (src_cksm);
				errno = sav_errno;
				return (-1);
			}
			if (filestatuses[0].status != 0) {
				gfal_errmsg (errbuf, errbufsz, GFAL_ERRLEVEL_ERROR, "%s: %s", filestatuses[0].surl, strerror (filestatuses[0].status));
				free (src_req);
				gfal_file_free (gfile_src);
				gfal_internal_free (dst_gfal);
				if (src_cksm) free (src_cksm);
				errno = filestatuses[0].status;
				return (-1);
			}
			strcpy (dest_turl, filestatuses[0].turl);

			if (gfal_set_xfer_running (dst_gfal, errbuf, errbufsz) < 0) {
				sav_errno = errno;
				free (src_req);
				gfal_abortrequest (dst_gfal, errmsg, GFAL_ERRMSG_LEN);
				gfal_deletesurls (dst_gfal, errmsg, GFAL_ERRMSG_LEN);
				gfal_file_free (gfile_src);
				gfal_internal_free (dst_gfal);
				if (src_cksm) free (src_cksm);
				errno = sav_errno;
				return (-1);
			}
			if ((n = lcg_gfal_functions.call_gfal_get_results (dst_gfal, &filestatuses)) < 0 || filestatuses == NULL) {
				sav_errno = errno;
				gfal_errmsg (errbuf, errbufsz, GFAL_ERRLEVEL_ERROR, "Internal error");
				free (src_req);
				gfal_abortrequest (dst_gfal, errmsg, GFAL_ERRMSG_LEN);
				gfal_deletesurls (dst_gfal, errmsg, GFAL_ERRMSG_LEN);
				gfal_file_free (gfile_src);
				gfal_internal_free (dst_gfal);
				if (src_cksm) free (src_cksm);
				errno = sav_errno;
				return (-1);
			}
			if (filestatuses[0].status != 0) {
				gfal_errmsg (errbuf, errbufsz, GFAL_ERRLEVEL_ERROR, "%s: %s",
						filestatuses[0].surl, strerror (filestatuses[0].status));
				free (src_req);
				gfal_abortrequest (dst_gfal, errmsg, GFAL_ERRMSG_LEN);
				gfal_deletesurls (dst_gfal, errmsg, GFAL_ERRMSG_LEN);
				gfal_file_free (gfile_src);
				gfal_internal_free (dst_gfal);
				if (src_cksm) free (src_cksm);
				errno = filestatuses[0].status;
				return (-1);
			}
		}

		if (bool_issurlok && cksmtype != GFAL_CKSM_NONE && src_cksm == NULL) {
			*errmsg = 0;
			if (!(bool_issurlok = gridftp_checksum (gfile_src->turl, cksmtype, &src_cksm, 1,
					errmsg, GFAL_ERRMSG_LEN, gfal_get_timeout_sendreceive ()) == 0 && src_cksm)) {
				gfal_file_set_turl_error (gfile_src, sav_errno, errmsg);
				if (gfile_src->gobj != NULL)
				    lcg_gfal_functions.call_gfal_set_xfer_done (gfile_src->gobj, errmsg, GFAL_ERRMSG_LEN);
			}
		}

		if (bool_issurlok) {
			char tmpbuf[21];
			gfal_errmsg (NULL, 0, GFAL_ERRLEVEL_WARN, "Source URL: %s", gfile_src->file);
			sprintf (tmpbuf, GFAL_LONG64_FORMAT, filesize);
			gfal_errmsg (NULL, 0, GFAL_ERRLEVEL_WARN, "File size: %s", tmpbuf);
			gfal_errmsg (NULL, 0, GFAL_ERRLEVEL_WARN, "Destination specified: %s", dest_file);
			gfal_errmsg (NULL, 0, GFAL_ERRLEVEL_WARN, "Source URL for copy: %s", gfile_src->turl);
			gfal_errmsg (NULL, 0, GFAL_ERRLEVEL_WARN, "Destination URL: %s", dest_turl);
			gfal_errmsg (NULL, 0, GFAL_ERRLEVEL_WARN, "# streams: %d", nbstreams);

			if (verbose)
				tm_start = times (&tmp_tms);

			/* copy file */

			*errmsg = 0;
			if (!(bool_issurlok = lcg_globus_functions.call_globus_copyfilex (gfile_src->turl, dest_turl, nbstreams, errmsg,
							GFAL_ERRMSG_LEN, gfal_get_timeout_sendreceive (), verbose) == 0)) {
				sav_errno = errno;
				if (sav_errno == ECANCELED) {
					gfal_errmsg (errbuf, errbufsz, GFAL_ERRLEVEL_ERROR, "Copy Cancelled...");

					if (gfile_src->gobj != NULL)
					    lcg_gfal_functions.call_gfal_set_xfer_done (gfile_src->gobj, errmsg, GFAL_ERRMSG_LEN);

					gfal_abortrequest (dst_gfal, errmsg, GFAL_ERRMSG_LEN);
					gfal_deletesurls (dst_gfal, errmsg, GFAL_ERRMSG_LEN);
					free (src_req);
					gfal_file_free (gfile_src);
					gfal_internal_free (dst_gfal);
					if (src_cksm) free (src_cksm);
					errno = sav_errno;
					return (-1);
				}

				gfal_file_set_turl_error (gfile_src, sav_errno, errmsg);
				if (gfile_src->gobj != NULL)
				    lcg_gfal_functions.call_gfal_set_xfer_done (gfile_src->gobj, errmsg, GFAL_ERRMSG_LEN);
			} else
				gfal_errmsg (NULL, 0, GFAL_ERRLEVEL_WARN, "\nTransfer took %d ms",
						(int) ((times (&tmp_tms) - tm_start) / (sysconf (_SC_CLK_TCK) / 1000.)));
		}

	      if (!bool_issurlok) {
            if (lcg_gfal_functions.call_gfal_file_next_replica(gfile_src) < 0) {
                gfal_errmsg(errbuf, errbufsz, GFAL_ERRLEVEL_ERROR, gfile_src->errmsg);
                errno = gfile_src->errcode;
                goto FUNCTION_ERROR_EXIT_WITH_CLEANUP;
            }
        }
	}
	free (src_req);

	if (gfile_src->errcode) {
		sav_errno = gfile_src->errcode;
		gfal_errmsg (errbuf, errbufsz, GFAL_ERRLEVEL_ERROR, "%s: %s", gfile_src->file,
				gfile_src->errmsg && gfile_src->errmsg[0] ? gfile_src->errmsg : strerror (sav_errno));
		gfal_abortrequest (dst_gfal, errmsg, GFAL_ERRMSG_LEN);
		gfal_deletesurls (dst_gfal, errmsg, GFAL_ERRMSG_LEN);
		gfal_file_free (gfile_src);
		gfal_internal_free (dst_gfal);
		if (src_cksm) free (src_cksm);
		errno = sav_errno;
		return (-1);
	}

	if (gfile_src->gobj != NULL)
	    lcg_gfal_functions.call_gfal_set_xfer_done (gfile_src->gobj, errbuf, errbufsz);

	sav_errno = 0;

	if (lcg_gfal_functions.call_gfal_set_xfer_done (dst_gfal, errbuf, errbufsz) < 0)
		sav_errno = errno;
	if (!sav_errno && ((n = lcg_gfal_functions.call_gfal_get_results (dst_gfal, &filestatuses)) < 0 || filestatuses == NULL)) {
		sav_errno = errno;
		gfal_errmsg (errbuf, errbufsz, GFAL_ERRLEVEL_ERROR, "Internal error");
	}
	if (!sav_errno && filestatuses[0].status != 0) {
		sav_errno = filestatuses[0].status;
		gfal_errmsg (errbuf, errbufsz, GFAL_ERRLEVEL_ERROR, "%s: %s", filestatuses[0].surl, strerror (filestatuses[0].status));
	}

	sav_errno = 0;

	if (cksmtype != GFAL_CKSM_NONE) {
		lcg_checksum *checksums = NULL;

		lcg_get_checksum_surls (dst_gfal, 0, cksmtype, &checksums, errbuf, errbufsz);

		if (checksums == NULL) {
			sav_errno = errno ? errno : EFAULT;
		}
		else if (checksums[0].status != 0) {
			sav_errno = checksums[0].status;
			if (checksums[0].explanation)
				gfal_errmsg (errbuf, errbufsz, GFAL_ERRLEVEL_ERROR, "%s", checksums[0].explanation);
			else
				gfal_errmsg (errbuf, errbufsz, GFAL_ERRLEVEL_ERROR, "[LCG-UTIL][lcg_get_checksum_surls][] %s: %s",
						checksums[0].file, strerror (checksums[0].status));
		}
		else if (checksums[0].checksum == NULL) {
			sav_errno = EFAULT;
		}
		else
			dst_cksm = checksums[0].checksum;

		if (checksums) {
			if (checksums[0].file) free (checksums[0].file);
			if (checksums[0].explanation) free (checksums[0].explanation);
			free (checksums);
		}

        lcgutil_str_tolower(src_cksm);
        lcgutil_str_tolower(dst_cksm);

		if (sav_errno == 0) {
			if (strcmp (src_cksm, dst_cksm)) {
				/* checksums are different!!! */
				sav_errno = EAGAIN;
				gfal_deletesurls (dst_gfal, errbuf, errbufsz);
				gfal_errmsg (errbuf, errbufsz, GFAL_ERRLEVEL_ERROR,
						"[LCG-UTIL][lcg_rep][] Destination may be corrupted: Source checksum (%s) != Destination checksum (%s)",
						src_cksm, dst_cksm);
				gfal_errmsg (errbuf, errbufsz, GFAL_ERRLEVEL_ERROR,
						"[LCG-UTIL][lcg_cp4][] %s has been DELETED, please try again (temporary network problem?)", dst_gfal->surls[0]);
			}
			else
				gfal_errmsg (errbuf, errbufsz, GFAL_ERRLEVEL_WARN,
						"[INFO] Source and destination checksums are matching: %s", src_cksm);
		}

		free (src_cksm);
		free (dst_cksm);
	}

	gfal_internal_free (dst_gfal);

	if (!sav_errno) {
		/* register file in file catalog */
		if (lcg_gfal_functions.call_gfal_register_file (gfile_src->lfn, gfile_src->guid, dest_surl_cat, 0, filesize, 0, errbuf, errbufsz) != 0) {
			sav_errno = errno;
			gfal_errmsg (NULL, 0, GFAL_ERRLEVEL_ERROR, "Could not register in Catalog the URL %s", dest_surl_cat);
			gfal_errmsg (NULL, 0, GFAL_ERRLEVEL_ERROR, "When the problem will be solved, it will have to be registered by hand!");
		} else
			gfal_errmsg (NULL, 0, GFAL_ERRLEVEL_WARN, "Destination URL registered in file catalog: %s", dest_surl_cat);
	}

	gfal_file_free (gfile_src);
	errno = sav_errno;
	return (sav_errno ? -1 : 0);

FUNCTION_ERROR_EXIT_WITH_CLEANUP:
    sav_errno = errno;
    if (gfile_src->gobj != NULL)
        lcg_gfal_functions.call_gfal_set_xfer_done(gfile_src->gobj, errmsg, GFAL_ERRMSG_LEN);
    gfal_abortrequest(dst_gfal, errmsg, GFAL_ERRMSG_LEN);
    gfal_deletesurls(dst_gfal, errmsg, GFAL_ERRMSG_LEN);
    free(src_req);
    gfal_file_free(gfile_src);
    gfal_internal_free(dst_gfal);
    if (src_cksm)
        free(src_cksm);
    errno = sav_errno;
    return (-1);
}
