/*
 * Copyright (c) Members of the EGEE Collaboration. 2004.
 * See http://www.eu-egee.org/partners/ for details on the copyright holders.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *     http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#ifndef lint
static char sccsid[] = "@(#)$RCSfile: lcg_uf.c,v $ $Revision: 1.23 $ $Date: 2008/12/18 19:58:08 $ CERN Jean-Philippe Baud";
#endif /* not lint */

/*	lcg_uf - unregisters in the LRC a file residing on an SE
 *	If it was the last or only valid replica, the entries corresponding to
 *	this GUID are also removed from the RMC.
 */
#include <stdio.h>
#include <errno.h>
#include <string.h>
#include <stdlib.h>
#include "gfal_api.h"
#include "lcg_util.h"

lcg_uf (char *surl, char *guid, char *vo, char *conf_file, int insecure)
{
	return (lcg_uf4 (surl, guid, 0, vo, 0, NULL, 0));
}

lcg_ufx (char *surl, char *guid, char *vo, char *conf_file, int insecure,
	char *errbuf, int errbufsz)
{
	return (lcg_uf4 (surl, guid, 0, vo, 0, errbuf, errbufsz));
}

lcg_uf3 (char *surl, char *guid, char *vo, char *conf_file, int insecure,
	int verbose, char *errbuf, int errbufsz)
{
	return (lcg_uf4 (surl, guid, 0, vo, verbose, errbuf, errbufsz));
}

lcg_uf4 (char *surl, char *guid, int nobdii, char *vo, int verbose, char *errbuf, int errbufsz)
{
	int rc = 0;
	int *results;

	gfal_set_verbose (verbose);

	if (surl == NULL || guid == NULL) {
		errno = EFAULT;
		return (-1);
	}
	if (vo && vo[0]) {
		if (strlen (vo) > GFAL_VO_MAXLEN) {
			gfal_errmsg (errbuf, errbufsz, GFAL_ERRLEVEL_ERROR, "%s: VO name too long", vo);
			errno = ENAMETOOLONG;
			return (-1);
		}
		if (gfal_set_vo (vo) < 0)
			return (-1);
	}

	gfal_set_nobdii (nobdii);

	if (strncmp (guid, "guid:", 5) == 0)
		guid += 5;	/* strip off guid prefix */

	/* unregister file */
	if (gfal_unregister_pfns (1, (const char **) &guid, (const char **) &surl,
				&results, errbuf, errbufsz) < 0 || !results)
		return (-1);

	rc = results[0];
	free (results);

	errno = rc;
	return (rc == 0 ? 0 : -1);
}
